
:root {
    --gammapy-primary: rgb(224, 10, 52);
    --gammapy-secondary: #1e254a;
    --gammapy-lightgray: #f2f2f2;;
}

/*
temporal fix for sphinx v2 and sphinx_rtd_theme
https://github.com/readthedocs/sphinx_rtd_theme/issues/766
*/
.classifier:before {
    font-style: normal;
    margin: 0.2em;
    content: ":";
}

/* sidebar header (and topbar for mobile) */
 .wy-side-nav-search, .wy-nav-top {
   background: var(--gammapy-secondary);
 }

/* sidebar */
 .wy-nav-side {
   background: var(--gammapy-secondary);
}

/* links outside nav */
.rst-content a:link, .rst-content a:visited {
   color: var(--gammapy-primary);
   text-decoration: none;
 }

/* admonitions only */
.admonition.note {
   background: var(--gammapy-lightgray);
}

/* font size smaller and striped cols in notebook tables */
.output_area {
    font-size: 12px;
}
.output_area tr:nth-child(even) {background-color: var(--gammapy-lightgray);}
th {
    background-color: #f2f2f2;
    padding-right: 8px;
    text-align: left;
}

/* font size smaller and not-stripped cols in API tables */
table.docutils {
    font-size: 14px;
}
.rst-content table.docutils:not(.field-list) tr:nth-child(2n-1) td,
.wy-table-striped tr:nth-child(2n-1) td,
.wy-table-odd td {
     background-color: #fff;
}
